


global QuixstreamServer QuixstreamPort RxLength EventOccurred UdpPacket


set EventOccurred 0

variable RxLength
variable UdpPacket


global StatusAddress DataAddress ArmAddress TraceAddress QuixTraceLength
set StatusAddress 0
set ArmAddress 6
set DataAddress 512
set TraceAddress 8192
set QuixTraceLength 512


proc MyLog {s} {
    global MIDAS_QServer_Verbose 
    if {$MIDAS_QServer_Verbose} {puts stdout "$s\n"}
}

proc Event {sock} {
    global EventOccurred UdpPacket RxLength

    MyLog "Event occurred - reading $RxLength"

    set required $RxLength
    set UdpPacket ""

    set tries 100

    while {$required > 0} {

        update idletasks

        set z [catch {read $sock $required} data]
        if {$z != 0} {
            MyLog "Error reading from network: $data"
            return -code error -errorcode "Error reading from network: $data" -errorinfo "Error detected in procedure Event"
        }

        set peer [fconfigure $sock -peer]
        MyLog "received [string length $data] bytes from $peer"

        set required [expr $required - [string length $data]]
        append UdpPacket $data

        incr tries -1
        if {$tries == 0} {
            MyLog "Timeout reading from network"
            return -code error -errorcode "Timeout reading from network" -errorinfo "Error detected in procedure Event"
        }

        if {[expr ($tries/20)*20] == $tries} {after 10}
    }

    set EventOccurred 1
}

proc Flush {sock} {

    MyLog "Flushing $sock"

    set tries 25

    while {$tries > 0} {

        update idletasks

        set z [catch {read $sock 64} data]
        if {$z != 0} {
            MyLog "Error reading from network: $data"
            return -code error -errorcode "Error reading from network: $data" -errorinfo "Error detected in procedure Event"
        }

        set peer [fconfigure $sock -peer]
        MyLog "received [string length $data] bytes from $peer"

        if {[string length $data] == 0} {return}    ;# no data - flush complete

        incr tries -1
        if {$tries == 0} {
            MyLog "Unable to flush network"
            return -code error -errorcode "Unable to flush network" -errorinfo "Error detected in procedure Event"
        }
    }
}


proc QServer_Init {} {
    global sockin sockout QuixstreamServer QuixstreamPort RxLength EventOccurred UdpPacket

    if {[info exists sockin] == 0} {set sockin 0}    ;#  define if it does not exist
    if {$sockin != 0}  {catch {close $sockin}; set sockin 0}            ;#  tidy up from previous aborted run

    if {[info exists sockout] == 0} {set sockout 0}    ;#  define if it does not exist
    if {$sockout != 0} {catch {close $sockout}; set sockout 0}          ;#  tidy up from previous aborted run

#    open UDP sockets

    set sockin [udp_open 5011]    ;#  open socket using port 5011
    MyLog "obtained socket $sockin"

    fconfigure $sockin -blocking yes  -buffering none -translation binary
    fileevent $sockin readable [list Event $sockin]

    set sockout [udp_open 5012]   ;#  open socket using port 5012
    MyLog "obtained socket $sockout"

    fconfigure $sockout -remote [list $QuixstreamServer $QuixstreamPort] -blocking yes  -buffering none -translation {binary binary} -encoding binary
    MyLog "configured socket for  host $QuixstreamServer port $QuixstreamPort"

#    configure fpga ethernet connection

    set command [list 0 0 0 0 0 0]
    set z [catch {puts -nonewline $sockout [binary format c* $command]} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Init (1) with errors: $m"
        return -code error -errorcode "abandoned QServer_Init (1) with errors: $m" -errorinfo "QServer_Init: error during configure fpga ethernet connection"
    }

    set RxLength 6; set EventOccurred 0
#    vwait EventOccurred
    Event $sockin

    set z [catch {binary scan $UdpPacket c* response} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Init (2) with errors: $m"
        return -code error -errorcode "abandoned QServer_Init (2) with errors: $m" -errorinfo "QServer_Init: error during configure fpga ethernet connection"
    }

    if {$z == 0 && $m == 0} {
        MyLog "abandoned QServer_Init (3) - null response"
        return -code error -errorcode "abandoned QServer_Init (3): null response" -errorinfo "QServer_Init: error during configure fpga ethernet connection"
    }

    if {[first $response] != 0} {    ;#   check that response is for initialiase (command=0)
        MyLog "abandoned QServer_Init (4) - bad response ($response)"
        return -code error -errorcode "abandoned QServer_Init (4): bad response" -errorinfo "QServer_Init: error during configure fpga ethernet connection"

    }

    MyLog "QServer_Init: Configured"

    return 0
}

proc QServer_Exit {} {
    global sockin  sockout

    if {[info exists sockin] == 0} {set sockin 0}    ;#  define if it does not exist
    if {$sockin != 0}  {catch {close $sockin}; set sockin 0}            ;#  tidy up from previous aborted run

    if {[info exists sockout] == 0} {set sockout 0}    ;#  define if it does not exist
    if {$sockout != 0} {catch {close $sockout}; set sockout 0}          ;#  tidy up from previous aborted run

   return 0

}


proc QServer_Read {Address Count} {
    global RxLength EventOccurred sockout sockin UdpPacket

#    data is A16 D32:    Count is the number of D32 data items

    set length [expr $Count * 2]     ;# number of 16 bit integers - must be a multiple of 2 between 2 and 16K
    set address $Address             ;# base address - must be a multiple of 2 between 2 and 16K

    set lengthH  [expr $length >> 8]    ;#  msb
    set lengthL  [expr $length & 255]  ;#  lsb

    set addressH  [expr $address >> 8]    ;#  msb
    set addressL  [expr $address & 255]  ;#  lsb

    catch {Flush $sockin}    ;#  flush the socket fifo
    
#    read fpga block

     set command [list 2 0 $lengthL $lengthH $addressL $addressH]
     set z [catch {puts -nonewline $sockout [binary format c* $command]} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Read (1) with errors: $m"
        return -code error -errorcode "abandoned QServer_Read (1) with errors: $m" -errorinfo "QServer_Read: error during read fpga block"
    }

     set RxLength 6; set EventOccurred 0
#     vwait EventOccurred
     Event $sockin

     set z [catch {binary scan $UdpPacket c* response} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Read (2) with errors: $m"
        return -code error -errorcode "abandoned QServer_Read 2) with errors: $m" -errorinfo "QServer_Read: error during read fpga block"
    }

    if {$z == 0 && $m == 0} {
        MyLog "abandoned QServer_Read (3) - null response"
        return -code error -errorcode "abandoned QServer_Read (3): null response" -errorinfo "QServer_Read: error during read fpga block"
    }

    if {[first $response] != 2} {    ;#   check that response is for read fpga block (command=2)
        MyLog "abandoned QServer_Read (4) - bad response ($response)"
        return -code error -errorcode "abandoned QServer_Read (4): bad response" -errorinfo "QServer_Read: error during read fpga block"

    }

#    read server block

    set command [list 3 0 $lengthL $lengthH $addressL $addressH]
    set z [catch {puts -nonewline $sockout [binary format c* $command]} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Read (5) with errors: $m"
        return -code error -errorcode "abandoned QServer_Read (5) with errors: $m" -errorinfo "QServer_Read: error during read server block"
    }

    set RxLength [expr $length * 2]; set EventOccurred 0
#    vwait EventOccurred
     Event $sockin

    set z [catch {binary scan $UdpPacket c* response} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Read (6) with errors: $m"
        return -code error -errorcode "abandoned QServer_Read (6) with errors: $m" -errorinfo "QServer_Read: error during read server block"
    }

    if {$z == 0 && [llength $response] == 0} {
        MyLog "abandoned QServer_Read (7) - null response"
        return -code error -errorcode "abandoned QServer_Read (7): null response" -errorinfo "QServer_Read: error during read server block"
    }

    set data {}; for {set i 0} {$i < [llength $response]} {} {
        set b1 [expr [lindex $response $i] & 255]; incr i
        set b0 [expr [lindex $response $i] & 255]; incr i

        set b3 [expr [lindex $response $i] & 255]; incr i
        set b2 [expr [lindex $response $i] & 255]; incr i

        append data [format "  0x%08x" [expr ((($b3 << 8) + $b2 << 8) + $b1 << 8) + $b0]]
    }

    set RxLength 6; set EventOccurred 0
#    vwait EventOccurred
     Event $sockin

    set z [catch {binary scan $UdpPacket c* response} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Read (8) with errors: $m"
        return -code error -errorcode "abandoned QServer_Read (8) with errors: $m" -errorinfo "QServer_Read: error during read server block"
    }

    if {$z == 0 && $m == 0} {
        MyLog "abandoned QServer_Read (9) - null response"
        return -code error -errorcode "abandoned QServer_Read (9): null response" -errorinfo "QServer_Read: error during read server block"
    }

    if {[first $response] != 3} {    ;#   check that response is for read server block (command=3)
        MyLog "abandoned QServer_Read (10) - bad response ($response)"
        return -code error -errorcode "abandoned QServer_Read (10): bad response" -errorinfo "QServer_Read: error during read server block"

    }

MyLog "Finished OK"

   return $data
}


proc QServer_Write {Address Count Data} {
    global RxLength EventOccurred sockout sockin UdpPacket

#    data is A16 D32:    Count is the number of D32 data items

    set length [expr $Count * 2]  ;# number of 16 bit integers - must be a multiple of 2 between 2 and 16K
    set address $Address          ;# base address - must be a multiple of 2 between 2 and 16K

    set lengthH  [expr $length >> 8]    ;#  msb
    set lengthL  [expr $length & 255]  ;#  lsb

    set addressH  [expr $address >> 8]    ;#  msb
    set addressL  [expr $address & 255]  ;#  lsb

    set data {}; for {set i 0} {$i < $Count } {incr i} {
        lappend data [expr ([lindex $Data $i] >> 8) & 255]
        lappend data [expr [lindex $Data $i] & 255]

        lappend data [expr ([lindex $Data $i] >> 24) & 255]
        lappend data [expr ([lindex $Data $i] >> 16) & 255]
    }

    catch {Flush $sockin}    ;#  flush the socket fifo

#    write server block

    set command [list 4 0 $lengthL $lengthH $addressL $addressH]
    set z [catch {puts -nonewline $sockout [binary format c* $command]} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Write (1) with errors: $m"
        return -code error -errorcode "abandoned QServer_Write (1) with errors: $m" -errorinfo "QServer_Write: error during write server block"
    }

    set z [catch {puts -nonewline $sockout [binary format c* $data]} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Write (2) with errors: $m"
        return -code error -errorcode "abandoned QServer_Write (2) with errors: $m" -errorinfo "QServer_Write: error during write server block"
    }

    set RxLength 6; set EventOccurred 0
#    vwait EventOccurred
    Event $sockin

    set z [catch {binary scan $UdpPacket c* response} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Write (3) with errors: $m"
        return -code error -errorcode "abandoned QServer_Write (3) with errors: $m" -errorinfo "QServer_Write: error during write server block"
    }

    if {$z == 0 && $m == 0} {
        MyLog "abandoned QServer_Write (4) - null response"
        return -code error -errorcode "abandoned QServer_Write (4): null response" -errorinfo "QServer_Write: error during write server block"
    }

    if {[first $response] != 4} {    ;#   check that response is for write server block (command=4)
        MyLog "abandoned QServer_Write (5) - bad response ($response)"
        return -code error -errorcode "abandoned QServer_Write (5): bad response" -errorinfo "QServer_Write: error during write server block"

    }

#    write fpga block

     set command [list 1 0 $lengthL $lengthH $addressL $addressH]
     set z [catch {puts -nonewline $sockout [binary format c* $command]} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Write (6) with errors: $m"
        return -code error -errorcode "abandoned QServer_Write (6) with errors: $m" -errorinfo "QServer_Write: error during write fpga block"
    }

     set RxLength 6; set EventOccurred 0
#     vwait EventOccurred
     Event $sockin

     set z [catch {binary scan $UdpPacket c* response} m]
    if {$z != 0} then {
        MyLog "abandoned QServer_Write (7) with errors: $m"
        return -code error -errorcode "abandoned QServer_Write (7) with errors: $m" -errorinfo "QServer_Write: error during write fpga block"
    }

    if {$z == 0 && $m == 0} {
        MyLog "abandoned QServer_Write (8) - null response"
        return -code error -errorcode "abandoned QServer_Write (8): null response" -errorinfo "QServer_Write: error during write fpga block"
    }

    if {[first $response] != 1} {    ;#   check that response is for write fpga block (command=1)
        MyLog "abandoned QServer_Write (9) - bad response ($response)"
        return -code error -errorcode "abandoned QServer_Write (9): bad response" -errorinfo "QServer_Write: error during write fpga block"

    }

MyLog "Finished OK"

   return 0
}


proc QServer_GetEvent {NumChans TraceLength} {
    global StatusAddress DataAddress ArmAddress
    global TraceAddress QuixTraceLength

    set status [QServer_Read $StatusAddress 1]

    if {[expr $status & 2] != 2} {return 0xffffffff}

    if {$NumChans == 0} then {set eventdata ""} else {set eventdata [QServer_Read $DataAddress $NumChans]}

    if {$TraceLength == 0} then {set tracedata ""} else {set tracedata [QServer_Read $TraceAddress $TraceLength]}

    set rc [QServer_Write $ArmAddress 1 0x1]

    return "$eventdata $tracedata"
}
